﻿// Класс-обертка для микроконтроллера ATMega16 и USB-драйвера V-USB (AVR-USB) Objective Development
// Подробности реализации - на сайте http://microsin.ru
// Copyright: (c) 2009 by Sergey Kukhtetskiy
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.InteropServices;
using System.Threading;

namespace AvrUsbDevice
{
    unsafe class ATMega16
    {
        #region Интерфейс с libusb

        #region Константы
        
        public const int LIBUSB_PATH_MAX = 512;
        const string LIBUSB_NATIVE_LIBRARY = "libusb0.dll";

        // Тип запроса (bmRequestType) формируется побитовым "или" 
        // констант из трех групп ниже.
        const byte USB_ENDPOINT_IN          = 0x80;
        const byte USB_ENDPOINT_OUT         = 0x00;

        const byte USB_TYPE_STANDARD        = (0x00 << 5);
        const byte USB_TYPE_CLASS           = (0x01 << 5);
        const byte USB_TYPE_VENDOR          = (0x02 << 5);
        const byte USB_TYPE_RESERVED        = (0x03 << 5);

        const byte USB_RECIP_DEVICE         = 0x00;
        const byte USB_RECIP_INTERFACE      = 0x01;
        const byte USB_RECIP_ENDPOINT       = 0x02;
        const byte USB_RECIP_OTHER          = 0x03;

        // Коды стандартных запросов (bRequest) с номерами 0 - 12 зарезервированы 
        // в спецификации USB поэтому все пользовательские будут начинаться с RQ_USER_BEG
        const byte RQ_USER_BEG              = 0x0C;

        const int  RQ_IO_READ               = 0x11;
        const int  RQ_IO_WRITE              = 0x12;
        #endregion

        #region Структуры

        /* Оригинальная структура из usb.h
        struct usb_endpoint_descriptor {
            unsigned char  bLength;
            unsigned char  bDescriptorType;
            unsigned char  bEndpointAddress;
            unsigned char  bmAttributes;
            unsigned short wMaxPacketSize;
            unsigned char  bInterval;
            unsigned char  bRefresh;
            unsigned char  bSynchAddress;
            unsigned char *extra;	// Extra descriptors
            int extralen;
        }; */
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi)]
        public struct usb_endpoint_descriptor
        {
            public byte bLength;
            public byte bDescriptorType;
            public byte bEndpointAddress;
            public byte bmAttributes;
            public byte wMaxPacketSize;
            public byte bInterval;
            public byte bRefresh;
            public byte bSynchAddress;
            public byte* extra;	// Extra descriptors
            public int extralen;
        };

        /* Оригинальная структура  из usb.h
        struct usb_interface_descriptor {
            unsigned char  bLength;
            unsigned char  bDescriptorType;
            unsigned char  bInterfaceNumber;
            unsigned char  bAlternateSetting;
            unsigned char  bNumEndpoints;
            unsigned char  bInterfaceClass;
            unsigned char  bInterfaceSubClass;
            unsigned char  bInterfaceProtocol;
            unsigned char  iInterface;
            struct usb_endpoint_descriptor *endpoint;
            unsigned char *extra;	// Extra descriptors 
            int extralen;
        }; */
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi)]
        public struct usb_interface_descriptor
        {
            public byte bLength;
            public byte bDescriptorType;
            public byte bInterfaceNumber;
            public byte bAlternateSetting;
            public byte bNumEndpoints;
            public byte bInterfaceClass;
            public byte bInterfaceSubClass;
            public byte bInterfaceProtocol;
            public byte iInterface;
            public usb_endpoint_descriptor* endpoint;
            public byte* extra;	// Extra descriptors 
            public int extralen;
        };

        /* Оригинальная структура  из usb.h
        struct usb_interface {
            struct usb_interface_descriptor *altsetting;
            int num_altsetting;
        }; */
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi)]
        public struct usb_interface
        {
            public usb_interface_descriptor* altsetting;
            public int num_altsetting;
        };

        /* Оригинальная структура  из usb.h
        struct usb_config_descriptor {
            unsigned char  bLength;
            unsigned char  bDescriptorType;
            unsigned short wTotalLength;
            unsigned char  bNumInterfaces;
            unsigned char  bConfigurationValue;
            unsigned char  iConfiguration;
            unsigned char  bmAttributes;
            unsigned char  MaxPower;
            struct usb_interface *interface;
            unsigned char *extra;	// Extra descriptors
            int extralen;
        }; */
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi)]
        public struct usb_config_descriptor
        {
            public byte bLength;
            public byte bDescriptorType;
            public ushort wTotalLength;
            public byte bNumInterfaces;
            public byte bConfigurationValue;
            public byte iConfiguration;
            public byte bmAttributes;
            public byte MaxPower;
            public usb_interface* pinterface;
            public byte* extra;	// Extra descriptors
            public int extralen;
        };

        /*  Оригинальная структура  из usb.h
        struct usb_device_descriptor {
            unsigned char  bLength;
            unsigned char  bDescriptorType;
            unsigned short bcdUSB;
            unsigned char  bDeviceClass;
            unsigned char  bDeviceSubClass;
            unsigned char  bDeviceProtocol;
            unsigned char  bMaxPacketSize0;
            unsigned short idVendor;
            unsigned short idProduct;
            unsigned short bcdDevice;
            unsigned char  iManufacturer;
            unsigned char  iProduct;
            unsigned char  iSerialNumber;
            unsigned char  bNumConfigurations;
        }; */
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi)]
        public struct usb_device_descriptor
        {
            public byte bLength;
            public byte bDescriptorType;
            public ushort bcdUSB;
            public byte bDeviceClass;
            public byte bDeviceSubClass;
            public byte bDeviceProtocol;
            public byte bMaxPacketSize0;
            public ushort idVendor;
            public ushort idProduct;
            public ushort bcdDevice;
            public byte iManufacturer;
            public byte iProduct;
            public byte iSerialNumber;
            public byte bNumConfigurations;
        };

        /*  Оригинальная структура  из usb.h
        struct usb_device 
        {
            struct usb_device *next, *prev;
            char filename[LIBUSB_PATH_MAX];
            struct usb_bus *bus;
            struct usb_device_descriptor descriptor;
            struct usb_config_descriptor *config;
            void *dev;		// Darwin support
            unsigned char devnum;
            unsigned char num_children;
            struct usb_device **children;
        }; */
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi)]
        public struct usb_device
        {
            public usb_device* next;
            public usb_device* prev;
            public fixed byte filename[LIBUSB_PATH_MAX];
            public usb_bus* bus;
            public usb_device_descriptor descriptor;
            public usb_config_descriptor* config;
            public IntPtr dev;		// Darwin support
            public byte devnum;
            public byte num_children;
            public IntPtr children;
        };

        /* Оригинальная структура  из usb.h
        struct usb_bus 
        {
            struct usb_bus *next, *prev;
            char dirname[LIBUSB_PATH_MAX];
            struct usb_device *devices;
            unsigned long location;
            struct usb_device *root_dev;
        };
        */
        [StructLayout(LayoutKind.Sequential, CharSet = CharSet.Ansi)]
        public struct usb_bus
        {
            public usb_bus* next;
            public usb_bus* prev;
            public fixed byte filename[LIBUSB_PATH_MAX];
            public usb_device* devices;
            public uint location;
            public usb_device* root_dev;
        };

        /* Оригинальная структура  из usbi.h
        struct usb_dev_handle {
            int fd;
            struct usb_bus *bus;
            struct usb_device *device;
            int config;
            int interface;
            int altsetting; 
            void *impl_info; // Added by RMT so implementations can store other per-open-device data
        };*/
        public struct usb_dev_handle
        {
            public int fd;
            public usb_bus* bus;
            public usb_device* device;
            public int config;
            public int iinterface;
            public int altsetting;
            public IntPtr impl_info; // Added by RMT so implementations can store other per-open-device data
        };
        #endregion

        #region Импорт функций libusb
        [DllImport(LIBUSB_NATIVE_LIBRARY)]
        public static extern void usb_init();
        [DllImport(LIBUSB_NATIVE_LIBRARY)]
        public static extern int usb_find_busses();
        [DllImport(LIBUSB_NATIVE_LIBRARY)]
        public static extern int usb_find_devices();
        [DllImport(LIBUSB_NATIVE_LIBRARY)]
        public static extern usb_bus* usb_get_busses();
        [DllImport(LIBUSB_NATIVE_LIBRARY)]
        public static extern usb_dev_handle* usb_open(usb_device* dev);
        [DllImport(LIBUSB_NATIVE_LIBRARY)]
        public static extern int usb_close(usb_dev_handle* dev);
        [DllImport(LIBUSB_NATIVE_LIBRARY)]
        public static extern int usb_control_msg(usb_dev_handle* dev, int requesttype, int request,
                                                 int value, int index, byte[] bytes, int size, int timeout);
        #endregion

        #endregion

        #region Инкапсуляция микроконтроллера ATMega16

        #region Адреса портов, регистров и пины микроконтроллера ATMega16
        // Значения этих констант взяты из файла WinAVR/avr/include/avr/iom16.h (definitions for ATmega16)
        public const byte aPINA     = 0x19;
        public const byte aDDRA     = 0x1A;
        public const byte aPORTA    = 0x1B;

        public const byte aPINB     = 0x16;
        public const byte aDDRB     = 0x17;
        public const byte aPORTB    = 0x18;

        public const byte aPINC     = 0x13;
        public const byte aDDRC     = 0x14;
        public const byte aPORTC    = 0x15;

        public const byte aPIND     = 0x10;
        public const byte aDDRD     = 0x11;
        public const byte aPORTD    = 0x12;

        public const byte PIN0      = 0;
        public const byte PIN1      = 1;
        public const byte PIN2      = 2;
        public const byte PIN3      = 3;
        public const byte PIN4      = 4;
        public const byte PIN5      = 5;
        public const byte PIN6      = 6;
        public const byte PIN7      = 7;

        public const byte aADCL     = 0x04;
        public const byte aADCH     = 0x05;

        public const byte aADCSRA   = 0x06;
        public const byte ADPS0     = 0;
        public const byte ADPS1     = 1;
        public const byte ADPS2     = 2;
        public const byte ADIE      = 3;
        public const byte ADIF      = 4;
        public const byte ADATE     = 5;
        public const byte ADSC      = 6;
        public const byte ADEN      = 7;

        public const byte aADMUX    = 0x07;
        public const byte MUX0      = 0;
        public const byte MUX1      = 1;
        public const byte MUX2      = 2;
        public const byte MUX3      = 3;
        public const byte MUX4      = 4;
        public const byte ADLAR     = 5;
        public const byte REFS0     = 6;
        public const byte REFS1     = 7;

        public const byte aICR1L    = 0x26;
        public const byte aICR1H    = 0x27;

        public const byte aOCR1BL   = 0x28;
        public const byte aOCR1BH   = 0x29;

        public const byte aOCR1AL   = 0x2A;
        public const byte aOCR1AH   = 0x2B;

        public const byte aTCNT1L   = 0x2C;
        public const byte aTCNT1H   = 0x2D;

        public const byte aTCCR1B   = 0x2E;
        public const byte CS10      = 0;
        public const byte CS11      = 1;
        public const byte CS12      = 2;
        public const byte WGM12     = 3;
        public const byte WGM13     = 4;
        public const byte ICES1     = 6;
        public const byte ICNC1     = 7;

        public const byte aTCCR1A   = 0x2F;
        public const byte WGM10     = 0;
        public const byte WGM11     = 1;
        public const byte FOC1B     = 2;
        public const byte FOC1A     = 3;
        public const byte COM1B0    = 4;
        public const byte COM1B1    = 5;
        public const byte COM1A0    = 6;
        public const byte COM1A1    = 7;
        #endregion

        #region Приватные члены
        private ushort vid, pid;
        private usb_dev_handle* handle = null;
        private usb_bus* bus;
        private usb_device* dev;
        private byte[] buffer = new byte[4];
        #endregion

        public ATMega16(ushort vid, ushort pid)
        {
            this.vid = vid; this.pid = pid;
            usb_init();
            int fb = usb_find_busses();
            int fd = usb_find_devices();
            byte[] buffer = new byte[4];

            for (bus = usb_get_busses(); bus != null && handle == null; bus = bus->next)
                for (dev = bus->devices; dev != null; dev = dev->next)
                    if (dev->descriptor.idVendor == vid && dev->descriptor.idProduct == pid)
                    {
                        handle = usb_open(dev);
                        break;
                    }
        }
        public bool IsOpen()
        {
            return handle != null;
        }
        public bool Close()
        {
            if (handle == null)
                return false;
            usb_close(handle);
            return true;
        }

        #region Работа с портами МК
        public byte DDRA
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aDDRA, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aDDRA, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte PORTA
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aPORTA, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aPORTA, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte PINA
        {
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aPINA, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte DDRB
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aDDRB, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aDDRB, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte PORTB
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aPORTB, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aPORTB, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte PINB
        {
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aPINB, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte DDRC
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aDDRC, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aDDRC, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte PORTC
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aPORTC, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aPORTC, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte PINC
        {
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aPINC, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte DDRD
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aDDRD, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aDDRD, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte PORTD
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aPORTD, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aPORTD, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte PIND
        {
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aPIND, buffer, 1, 5000);

                return buffer[0];
            }
        }
        #endregion 
        
        #region Работа с регистрами АЦП
        public byte ADCL    // Младшие биты кода АЦП
        {
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aADCL, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte ADCH    // Старшие биты кода АЦП
        {
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aADCH, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte ADCSRA  // Регистр статуса
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aADCSRA, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aADCSRA, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte ADMUX   // Регистр мультиплексора и типа источника опорного напряжения
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aADMUX, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aADMUX, buffer, 1, 5000);

                return buffer[0];
            }
        }
        #endregion

        #region Работа с регистрами TIMER1
        public byte ICR1L
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aICR1L, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aICR1L, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte ICR1H
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aICR1H, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aICR1H, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte OCR1BL
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aOCR1BL, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aOCR1BL, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte OCR1BH
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aOCR1BH, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aOCR1BH, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte OCR1AL
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aOCR1AL, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aOCR1AL, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte OCR1AH
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aOCR1AH, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aOCR1AH, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte TCNT1L
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aTCNT1L, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aTCNT1L, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte TCNT1H
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aTCNT1H, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aTCNT1H, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte TCCR1B
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aTCCR1B, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aTCCR1B, buffer, 1, 5000);

                return buffer[0];
            }
        }
        public byte TCCR1A
        {
            set
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT,
                                RQ_IO_WRITE, value, aTCCR1A, null, 0, 5000);
            }
            get
            {
                usb_control_msg(handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN,
                                RQ_IO_READ, 0, aTCCR1A, buffer, 1, 5000);

                return buffer[0];
            }
        }
        #endregion        

        #endregion
    }
}
